@@ -423,9 +423,7 @@ static Bool has_kernel_mode_setting(const struct pci_device *dev)
 			struct drm_mode_card_res res;
 
 			memset(&res, 0, sizeof(res));
-			if (drmIoctl(fd, DRM_IOCTL_MODE_GETRESOURCES, &res) == 0)
-				ret = res.count_crtcs != 0;
-			else
+			if (drmIoctl(fd, DRM_IOCTL_MODE_GETRESOURCES, &res))
 				ret = FALSE;
 		}
 		close(fd);
