@@ -38,6 +38,13 @@
 #define DBG_FORCE_UPLOAD 0
 #define DBG_NO_CPU_BO 0
 
+#define alphaless(format) PICT_FORMAT(PICT_FORMAT_BPP(format),		\
+				      PICT_FORMAT_TYPE(format),		\
+				      0,				\
+				      PICT_FORMAT_R(format),		\
+				      PICT_FORMAT_G(format),		\
+				      PICT_FORMAT_B(format))
+
 CARD32
 sna_format_for_depth(int depth)
 {
@@ -87,8 +94,8 @@ no_render_composite(struct sna *sna,
 	if (mask)
 		return false;
 
-	if (!is_gpu(dst->pDrawable) &&
-	    (src->pDrawable == NULL || !is_gpu(src->pDrawable)))
+	if (!is_gpu(sna, dst->pDrawable, PREFER_GPU_BLT) &&
+	    (src->pDrawable == NULL || !is_gpu(sna, src->pDrawable, PREFER_GPU_BLT)))
 		return false;
 
 	return sna_blt_composite(sna,
@@ -246,7 +253,7 @@ static void
 no_render_context_switch(struct kgem *kgem,
 			 int new_mode)
 {
-	if (!kgem->mode)
+	if (!kgem->nbatch)
 		return;
 
 	if (kgem_ring_is_idle(kgem, kgem->ring)) {
@@ -254,7 +261,6 @@ no_render_context_switch(struct kgem *kgem,
 		_kgem_submit(kgem);
 	}
 
-	(void)kgem;
 	(void)new_mode;
 }
 
@@ -280,7 +286,9 @@ void no_render_init(struct sna *sna)
 {
 	struct sna_render *render = &sna->render;
 
-	memset (render,0, sizeof (*render));
+	memset (render, 0, sizeof (*render));
+
+	render->prefer_gpu = PREFER_GPU_BLT;
 
 	render->vertices = render->vertex_data;
 	render->vertex_size = ARRAY_SIZE(render->vertex_data);
@@ -305,6 +313,8 @@ void no_render_init(struct sna *sna)
 	sna->kgem.expire = no_render_expire;
 	if (sna->kgem.has_blt)
 		sna->kgem.ring = KGEM_BLT;
+
+	sna_vertex_init(sna);
 }
 
 static struct kgem_bo *
@@ -339,6 +349,11 @@ use_cpu_bo(struct sna *sna, PixmapPtr pixmap, const BoxRec *box, bool blt)
 			     __FUNCTION__));
 			break;
 		default:
+			if (kgem_bo_is_busy(priv->gpu_bo)){
+				DBG(("%s: box is partially damaged on the CPU, and the GPU is busy\n",
+				     __FUNCTION__));
+				return NULL;
+			}
 			if (sna_damage_contains_box(priv->gpu_damage,
 						    box) != PIXMAN_REGION_OUT) {
 				DBG(("%s: box is damaged on the GPU\n",
@@ -448,7 +463,9 @@ move_to_gpu(PixmapPtr pixmap, const BoxRec *box, bool blt)
 
 	w = box->x2 - box->x1;
 	h = box->y2 - box->y1;
-	if (w == pixmap->drawable.width && h == pixmap->drawable.height) {
+	if (priv->cpu_bo && !priv->cpu_bo->flush) {
+		migrate = true;
+	} else if (w == pixmap->drawable.width && h == pixmap->drawable.height) {
 		migrate = priv->source_count++ > SOURCE_BIAS;
 
 		DBG(("%s: migrating whole pixmap (%dx%d) for source (%d,%d),(%d,%d), count %d? %d\n",
@@ -473,10 +490,19 @@ move_to_gpu(PixmapPtr pixmap, const BoxRec *box, bool blt)
 		migrate = count*w*h > pixmap->drawable.width * pixmap->drawable.height;
 	}
 
-	if (migrate && !sna_pixmap_force_to_gpu(pixmap,
-						blt ? MOVE_READ : MOVE_SOURCE_HINT | MOVE_READ))
+	if (!migrate)
 		return NULL;
 
+	if (blt) {
+		if (!sna_pixmap_move_area_to_gpu(pixmap, box,
+						 __MOVE_FORCE | MOVE_READ))
+			return NULL;
+	} else {
+		if (!sna_pixmap_move_to_gpu(pixmap,
+					    __MOVE_FORCE | MOVE_SOURCE_HINT | MOVE_READ))
+			return NULL;
+	}
+
 	return priv->gpu_bo;
 }
 
@@ -509,7 +535,7 @@ static struct kgem_bo *upload(struct sna *sna,
 			if (priv->ptr == NULL) /* uninitialised */
 				return NULL;
 			assert(priv->stride);
-			pixmap->devPrivate.ptr = priv->ptr;
+			pixmap->devPrivate.ptr = PTR(priv->ptr);
 			pixmap->devKind = priv->stride;
 		}
 	}
@@ -529,7 +555,6 @@ static struct kgem_bo *upload(struct sna *sna,
 		    channel->width  == pixmap->drawable.width &&
 		    channel->height == pixmap->drawable.height) {
 			assert(priv->gpu_damage == NULL);
-			assert(DAMAGE_IS_ALL(priv->cpu_damage));
 			assert(priv->gpu_bo == NULL);
 			kgem_proxy_bo_attach(bo, &priv->gpu_bo);
 		}
@@ -591,6 +616,10 @@ sna_render_pixmap_bo(struct sna *sna,
 		    !priv->cpu_bo->snoop && priv->cpu_bo->pitch < 4096) {
 			DBG(("%s: CPU all damaged\n", __FUNCTION__));
 			channel->bo = priv->cpu_bo;
+			if (priv->shm) {
+				assert(!priv->flush);
+				sna_add_flush_pixmap(sna, priv, priv->cpu_bo);
+			}
 			goto done;
 		}
 	}
@@ -877,6 +906,9 @@ sna_render_pixmap_partial(struct sna *sna,
 		DBG(("%s: tile size for tiling %d: %dx%d, size=%d\n",
 		     __FUNCTION__, bo->tiling, tile_width, tile_height, tile_size));
 
+		if (sna->kgem.gen < 033)
+			tile_width = bo->pitch;
+
 		/* Ensure we align to an even tile row */
 		box.y1 = box.y1 & ~(2*tile_height - 1);
 		box.y2 = ALIGN(box.y2, 2*tile_height);
@@ -979,11 +1011,14 @@ sna_render_picture_partial(struct sna *sna,
 	if (use_cpu_bo(sna, pixmap, &box, false)) {
 		bo = sna_pixmap(pixmap)->cpu_bo;
 	} else {
-		if (!sna_pixmap_force_to_gpu(pixmap,
-					     MOVE_READ | MOVE_SOURCE_HINT))
+		struct sna_pixmap *priv;
+
+		priv = sna_pixmap_force_to_gpu(pixmap,
+					       MOVE_READ | MOVE_SOURCE_HINT);
+		if (priv == NULL)
 			return 0;
 
-		bo = sna_pixmap(pixmap)->gpu_bo;
+		bo = priv->gpu_bo;
 	}
 
 	if (bo->pitch > sna->render.max_3d_pitch)
@@ -1159,13 +1194,11 @@ sna_render_picture_extract(struct sna *sna,
 						      pixmap->devKind,
 						      pixmap->drawable.bitsPerPixel);
 			if (bo != NULL &&
-			    pixmap->usage_hint == 0 &&
 			    box.x2 - box.x1 == pixmap->drawable.width &&
 			    box.y2 - box.y1 == pixmap->drawable.height) {
 				struct sna_pixmap *priv = sna_pixmap(pixmap);
 				if (priv) {
 					assert(priv->gpu_damage == NULL);
-					assert(DAMAGE_IS_ALL(priv->cpu_damage));
 					assert(priv->gpu_bo == NULL);
 					kgem_proxy_bo_attach(bo, &priv->gpu_bo);
 				}
@@ -1424,7 +1457,8 @@ sna_render_picture_approximate_gradient(struct sna *sna,
 	return -1;
 #endif
 
-	DBG(("%s: (%d, %d)x(%d, %d)\n", __FUNCTION__, x, y, w, h));
+	DBG(("%s: (%d, %d)x(%d, %d), dst=(%d, %d)\n",
+	     __FUNCTION__, x, y, w, h, dst_x, dst_y));
 
 	if (w2 == 0 || h2 == 0) {
 		DBG(("%s: fallback - unknown bounds\n", __FUNCTION__));
@@ -1435,6 +1469,7 @@ sna_render_picture_approximate_gradient(struct sna *sna,
 		return -1;
 	}
 
+	assert(channel->card_format == -1);
 	channel->pict_format = PIXMAN_a8r8g8b8;
 	channel->bo = kgem_create_buffer_2d(&sna->kgem,
 					    w2, h2, 32,
@@ -1450,6 +1485,7 @@ sna_render_picture_approximate_gradient(struct sna *sna,
 				       w2, h2, ptr, channel->bo->pitch);
 	if (!dst) {
 		kgem_bo_destroy(&sna->kgem, channel->bo);
+		channel->bo = NULL;
 		return 0;
 	}
 
@@ -1457,22 +1493,37 @@ sna_render_picture_approximate_gradient(struct sna *sna,
 	if (src == NULL) {
 		pixman_image_unref(dst);
 		kgem_bo_destroy(&sna->kgem, channel->bo);
+		channel->bo = NULL;
 		return 0;
 	}
+	DBG(("%s: source offset (%d, %d)\n", __FUNCTION__, dx, dy));
 
 	memset(&t, 0, sizeof(t));
 	t.matrix[0][0] = (w << 16) / w2;
+	t.matrix[0][2] = (x + dx) << 16;
 	t.matrix[1][1] = (h << 16) / h2;
+	t.matrix[1][2] = (y + dy) << 16;
 	t.matrix[2][2] = 1 << 16;
 	if (picture->transform)
 		pixman_transform_multiply(&t, picture->transform, &t);
+	DBG(("%s: applying transform [(%f, %f, %f), (%f, %f, %f), (%f, %f, %f)]\n",
+	     __FUNCTION__,
+	     pixman_fixed_to_double(t.matrix[0][0]),
+	     pixman_fixed_to_double(t.matrix[0][1]),
+	     pixman_fixed_to_double(t.matrix[0][2]),
+	     pixman_fixed_to_double(t.matrix[1][0]),
+	     pixman_fixed_to_double(t.matrix[1][1]),
+	     pixman_fixed_to_double(t.matrix[1][2]),
+	     pixman_fixed_to_double(t.matrix[2][0]),
+	     pixman_fixed_to_double(t.matrix[2][1]),
+	     pixman_fixed_to_double(t.matrix[2][2])));
 	pixman_image_set_transform(src, &t);
 
-	pixman_image_composite(PictOpSrc, src, NULL, dst,
-			       x + dx, y + dy,
-			       0, 0,
-			       0, 0,
-			       w2, h2);
+	sna_image_composite(PictOpSrc, src, NULL, dst,
+			    0, 0,
+			    0, 0,
+			    0, 0,
+			    w2, h2);
 	free_pixman_pict(picture, src);
 	pixman_image_unref(dst);
 
@@ -1521,7 +1572,8 @@ sna_render_picture_fixup(struct sna *sna,
 
 	if (picture->alphaMap) {
 		DBG(("%s: alphamap\n", __FUNCTION__));
-		if (is_gpu(picture->pDrawable) || is_gpu(picture->alphaMap->pDrawable)) {
+		if (is_gpu(sna, picture->pDrawable, PREFER_GPU_RENDER) ||
+		    is_gpu(sna, picture->alphaMap->pDrawable, PREFER_GPU_RENDER)) {
 			return sna_render_picture_flatten(sna, picture, channel,
 							  x, y, w, h, dst_x, dst_y);
 		}
@@ -1531,7 +1583,7 @@ sna_render_picture_fixup(struct sna *sna,
 
 	if (picture->filter == PictFilterConvolution) {
 		DBG(("%s: convolution\n", __FUNCTION__));
-		if (is_gpu(picture->pDrawable)) {
+		if (is_gpu(sna, picture->pDrawable, PREFER_GPU_RENDER)) {
 			return sna_render_picture_convolve(sna, picture, channel,
 							   x, y, w, h, dst_x, dst_y);
 		}
@@ -1544,10 +1596,6 @@ do_fixup:
 		channel->pict_format = PIXMAN_a8;
 	else
 		channel->pict_format = PIXMAN_a8r8g8b8;
-	if (channel->pict_format != picture->format) {
-		DBG(("%s: converting to %08x from %08x\n",
-		     __FUNCTION__, channel->pict_format, picture->format));
-	}
 
 	if (picture->pDrawable &&
 	    !sna_drawable_move_to_cpu(picture->pDrawable, MOVE_READ))
@@ -1564,8 +1612,10 @@ do_fixup:
 	}
 
 	/* Composite in the original format to preserve idiosyncracies */
-	if (picture->format == channel->pict_format)
-		dst = pixman_image_create_bits(picture->format,
+	if (!kgem_buffer_is_inplace(channel->bo) &&
+	    (picture->pDrawable == NULL ||
+	     alphaless(picture->format) == alphaless(channel->pict_format)))
+		dst = pixman_image_create_bits(channel->pict_format,
 					       w, h, ptr, channel->bo->pitch);
 	else
 		dst = pixman_image_create_bits(picture->format, w, h, NULL, 0);
@@ -1583,15 +1633,15 @@ do_fixup:
 
 	DBG(("%s: compositing tmp=(%d+%d, %d+%d)x(%d, %d)\n",
 	     __FUNCTION__, x, dx, y, dy, w, h));
-	pixman_image_composite(PictOpSrc, src, NULL, dst,
-			       x + dx, y + dy,
-			       0, 0,
-			       0, 0,
-			       w, h);
+	sna_image_composite(PictOpSrc, src, NULL, dst,
+			    x + dx, y + dy,
+			    0, 0,
+			    0, 0,
+			    w, h);
 	free_pixman_pict(picture, src);
 
 	/* Then convert to card format */
-	if (picture->format != channel->pict_format) {
+	if (pixman_image_get_data(dst) != ptr) {
 		DBG(("%s: performing post-conversion %08x->%08x (%d, %d)\n",
 		     __FUNCTION__,
 		     picture->format, channel->pict_format,
@@ -1684,13 +1734,13 @@ sna_render_picture_convert(struct sna *sna,
 	     pixmap->drawable.width,
 	     pixmap->drawable.height));
 
-	if (w == 0 || h == 0) {
+	if (w <= 0 || h <= 0) {
 		DBG(("%s: sample extents lie outside of source, using clear\n",
 		     __FUNCTION__));
 		return 0;
 	}
 
-	if (fixup_alpha && is_gpu(&pixmap->drawable)) {
+	if (fixup_alpha && is_gpu(sna, &pixmap->drawable, PREFER_GPU_RENDER)) {
 		ScreenPtr screen = pixmap->drawable.pScreen;
 		PixmapPtr tmp;
 		PicturePtr src, dst;
@@ -1705,7 +1755,9 @@ sna_render_picture_convert(struct sna *sna,
 						   PICT_FORMAT_B(picture->format));
 
 		DBG(("%s: converting to %08x from %08x using composite alpha-fixup\n",
-		     __FUNCTION__, picture->format));
+		     __FUNCTION__,
+		     (unsigned)channel->pict_format,
+		     (unsigned)picture->format));
 
 		tmp = screen->CreatePixmap(screen, w, h, pixmap->drawable.bitsPerPixel, 0);
 		if (tmp == NULL)
@@ -1726,7 +1778,7 @@ sna_render_picture_convert(struct sna *sna,
 						       pixmap->drawable.depth,
 						       picture->format),
 				    0, NULL, serverClient, &error);
-		if (dst == NULL) {
+		if (src == NULL) {
 			FreePicture(dst, 0);
 			screen->DestroyPixmap(tmp);
 			return 0;
@@ -1821,6 +1873,8 @@ sna_render_composite_redirect(struct sna *sna,
 	int bpp = op->dst.pixmap->drawable.bitsPerPixel;
 	struct kgem_bo *bo;
 
+	assert(t->real_bo == NULL);
+
 #if NO_REDIRECT
 	return false;
 #endif
@@ -1902,6 +1956,7 @@ sna_render_composite_redirect(struct sna *sna,
 			t->real_bo = op->dst.bo;
 			t->real_damage = op->damage;
 			if (op->damage) {
+				assert(!DAMAGE_IS_ALL(op->damage));
 				t->damage = sna_damage_create();
 				op->damage = &t->damage;
 			}
@@ -1918,6 +1973,7 @@ sna_render_composite_redirect(struct sna *sna,
 			}
 
 			assert(op->dst.bo != t->real_bo);
+			op->dst.bo->unique_id = kgem_get_unique_id(&sna->kgem);
 			op->dst.bo->pitch = t->real_bo->pitch;
 
 			op->dst.x -= box.x1;
@@ -1994,7 +2050,7 @@ sna_render_composite_redirect_done(struct sna *sna,
 		}
 		if (t->damage) {
 			DBG(("%s: combining damage (all? %d), offset=(%d, %d)\n",
-			     __FUNCTION__, DAMAGE_IS_ALL(t->damage),
+			     __FUNCTION__, (int)DAMAGE_IS_ALL(t->damage),
 			     t->box.x1, t->box.y1));
 			sna_damage_combine(t->real_damage,
 					   DAMAGE_PTR(t->damage),
